/*
   Trabajo de Fin de Grado
   Título: SISTEMA DE RIEGO AUTOMATIZADO PARA PEQUEÑAS INSTALACIONES
   Autora: Paula Zapatera Rodríguez
   Tutor: Ángel Eugenio Arranz Simón
   Departamento: Tecnología Electrónica
   Grado en Ingeniería en Electrónica Industrial y Automática
   Escuela de Ingenierías Industriales - Universidad de Valladolid
   Curso 2021 - 2022
*/

//////// Declaración de librerías:
///  (Se describen con más detalle en la memoria del proyecto).
#include <Wire.h> //Permite la comunicación por bus I2C
#include <OLED_I2C.h> //Pantalla OLED con comunicación I2C
#include <DHT.h> //Sensor de temperatura y humedad DHT11
#include <Servo.h> //Servomotor
//Control del reloj de tiempo real (módulo RTC)
#include <RTClib.h>
#include <SimpleTimer.h>

//////// Asignación de pines:
OLED  myOLED(SDA, SCL); //Asignación de pines para la pantalla
int sensorDHT = 7; //Pin digital 7 asignado al sensor DHT11
int sensorRes = A0; //Pin analógico A0 asigando al sensor de humedad resistivo
int sensorCap = A2; //Pin analógico A1 asigando al sensor de humedad capacitivo
int sensorHW = A1; //Pin analógico A1 asigando al sensor de nivel de agua
int bomba = 8; //Pin digital 8 asignado al relé que controla la bomba de agua
int pinServo = 9; //Pin digital 9 asignado al servomotor
const int redLED = 6; //Pin digital 6 asignado al LED rojo
const int greenLED = 10; //Pin digital 10 asignado al LED verde
const int yellLED = 11; //Pin digital 11 asignado al LED amarillo
//Declaración de los pines asignados a cada pulsador
int z = digitalRead(4); //Pulsador con función "atrás", pin digital 4
int y = digitalRead(3); //Pulsador con función "OK", pin digital 3
int x = digitalRead(2); //Pulsador con función "siguiente", pin digital 2
int w = digitalRead(5); //Pulsador para encender/apagar pantalla, pin digital 5

//////// Declaración de objetos:
RTC_DS3231 RTC; //Declaración del módulo RTC
#define DS3231_I2C_ADDRESS 0x68 //Dirección del módulo RTC
DHT dht (sensorDHT, DHT11); //Inicialización del sensor DHT11, declaración del pin y del tipo de DHT
Servo servo; //Declaración del servo

//////// Declaración de variables:
extern uint8_t SmallFont[]; //Fuente utilizada en la pantalla incluida en la librería OLED_I2C
int temp, humDHT; //Variables que guardan la temperatura y la humedad leídas por el sensor DHT11
int humedad; //Variable que guarda el valor de la humedad de la tierra
int nivel; //Variable para guardar el valor del nivel de agua del depósito
int i  = 13; //Posicion en y del cursor ">" en la pantalla
int page = 0; //Página por defecto
//Variables que guardan el estado de activación de cada elemento:
boolean releON = false, autoON = false;
boolean selecDown = true, selecBack = true, selecOk = true, selecSleep = true, selecRele = true;
boolean currentSleep = true; //Guarda el estado actual de la pantalla (encendida/apagada)
int pulsoMIN = 550; //Pulso mínimo del servo correspondiente al ángulo mínimo (0º)
int pulsoMAX = 2300; //Pulso máximo del servo correspondiente al ángulo máximo (180º)
//Variables para calcular el tiempo que ha pasado desde el inicio del programa
long currentMillis = 0;
long previousMillis = 0;
long autoMillis = 0;
boolean regando = false; //Indica si se está regando o no
// Horas de riego automático
int h_riego1 = 8; //Hora de comienzo del primer riego
int m_riego1 = 0; //Minuto de comienzo del primer riego
int h_riego2 = 21; //Hora de comienzo del segundo riego
int m_riego2 = 0; //Minuto de comienzo del segundo riego
int prox_riego = 0; //Guarda el tiempo que queda hasta el próximo riego programado

//Función para convertir datos en formato BCD a decimal:
byte bcdToDec(byte val) {
  return ( (val / 16 * 10) + (val % 16) );
}


//////// FUNCIÓN SETUP:
void setup() {

  myOLED.begin(); //Inicio de la pantalla OLED
  myOLED.clrScr(); //Limpiar la pantalla al inicio
  myOLED.setFont(SmallFont);//Establecer tipo de letra en la pantalla, cada caracter ocupa 6x8px

  RTC.begin(); //Inicio del módulo RTC
  //ajustar el reloj a la hora y fecha del ordenador
  //comentar esta línea después de cargar por primera vez el programa en la tarjeta Arduino
  RTC.adjust(DateTime(F(__DATE__), F(__TIME__)));

  dht.begin(); //Inicio del sensor de humedad DHT11

  pinMode(bomba, OUTPUT); //Asignar el pin 8 como salida
  digitalWrite(bomba, HIGH); //Inicia el relé apagado

  servo.attach(pinServo, pulsoMIN, pulsoMAX); //Declaración de parámetros del servomotor
  servo.write(180); //Inicia el deposito cerrado

  myOLED.sleepMode(SLEEP_ON); //La pantalla empieza apagada al cargar el programa
  myOLED.update();
}


//////// FUNCIÓN LOOP:
void loop() {

  //Llamada de funciones
  lecturas();
  RiegoAuto();
  sleep();
  back();

  //// Página 0: muestra la hora y fecha actuales y permite acceder al menú principal
  // mediante el botón "OK".
  if (page == 0) {

    myOLED.clrScr();
    myOLED.invertText(true);
    myOLED.print("        MENU         ", CENTER, 57);
    myOLED.invertText(false);
    pagina0(); //Llamada a la función de pág. 0
    myOLED.update();

    if (ok())
      page = 1; //Si se pulsa "OK" estando en la pág. 0, se accede a la pág. 1
  }

  //// Página 1: menú principal.
  if (page == 1) {

    myOLED.clrScr();
    myOLED.invertText(true);
    myOLED.print("        MENU         ", CENTER, 0);
    myOLED.invertText(false);
    myOLED.print("RIEGO AUTO: ", 10, 13);
    myOLED.print("RIEGO MANUAL: ", 10, 27);
    myOLED.print("DATOS ACTUALES", 10, 41);
    myOLED.print("ESTADO DEPOSITO", 10, 55);
    //Indicador de ON/OFF del modo automático
    if (autoON == true && page == 1)
      myOLED.print(" ON", 85, 13);
    if (autoON == false && page == 1)
      myOLED.print(" OFF", 85, 13);
    //Indicador de ON/OFF del modo manual
    if (releON == true && page == 1)
      myOLED.print(" ON", 85, 27);
    if (releON == false && page == 1)
      myOLED.print(" OFF", 85, 27);
    //Situación del cursor indicador
    myOLED.print(">", 0, i);
    myOLED.update();
    down(4); //Navegar por 4 opciones distintas
    riegoManual(); //Llamada a la función de riego manual
  }


  //Si el cursor está en la posición y = 55 y se pulsa "OK" se accede a la pág. 2
  if (i == 55 && ok()) {
    page = 2;
    i = 15;
  }

  //// Página 2: información sobre el nivel del depósito de agua.
  // Opción de abrir y cerrar manualmente el depósito.
  if (page == 2) {
    nivelAgua();
    servoDep();
    myOLED.clrScr();
    myOLED.invertText(true);
    myOLED.print("    ESTADO DEPOSITO    ", CENTER, 0);
    myOLED.invertText(false);
    myOLED.print("Nivel de agua: ", 10, 15);
    myOLED.printNumI(nivel, 106, 15);
    myOLED.print("%", 124, 15);
    myOLED.print("Abrir deposito", 10, 31);
    myOLED.print("Cerrar deposito", 10, 47);
    //Al entrar en la pág. 2 el cursor se sitúa en la posición y = 15+16:
    myOLED.print(">", 0, i + 16);
    myOLED.update();
    down(2); //En esta pág. las opciones seleccionables son 2
  }

  //Si el cursor está en y = 41, en la pág. 1 y se pulsa "OK" se accede a la pág. 3
  if (i == 41 && page == 1 && ok() ) {
    page = 3;
  }

  //// Página 3: información sobre datos actuales de humedad y Tª.
  if (page == 3 ) {

    leerDHT();
    sensorHum();
    myOLED.clrScr();
    myOLED.invertText(true);
    myOLED.print("    DATOS ACTUALES    ", CENTER, 0);
    myOLED.invertText(false);
    myOLED.print("T AMBIENTE: ", 10, 15);
    myOLED.printNumI(temp, 82, 15);
    myOLED.print("*C", 100, 15);
    myOLED.print("HUMEDAD AMB: ", 10, 31);
    myOLED.printNumI(humDHT, 86, 31);
    myOLED.print("%", 106, 31);
    myOLED.print("HUM TIERRA: ", 10, 47);
    myOLED.printNumI(humedad, 82, 47);
    myOLED.print("%", 106, 47);
    myOLED.print("->", 115, 55); //Se imprime una flecha -> para indicar la siguiente pág.
    myOLED.update();
  }

  // Estando en la pág 3. y pulsando "SIGUIENTE" se accede a la pág 4.
  if (page == 3 && x == 1)
    page = 4;

  //// Página 4: información que variará en función de las condiciones en las
  // que se encuentra el sistema en el momento de consulta de los datos.
  if (page == 4) {
    myOLED.clrScr();
    myOLED.invertText(true);
    myOLED.print("    DATOS ACTUALES    ", CENTER, 0);
    myOLED.invertText(false);
    pagina4(); //Llamada a la función que define la pág. 4
    myOLED.print("<-", 10, 55); //Se imprime una flecha -> para indicar pág. anterior
    myOLED.update();
  }
}


////FUNCIÓN readDS3231time:
void readDS3231time(byte *second, byte *minute, byte *hour, byte *dayOfWeek, byte *dayOfMonth, byte *month, byte *year) {
  Wire.beginTransmission(DS3231_I2C_ADDRESS); //Empieza transmisión a RTC
  Wire.write(0); //El puntero de registro se pone a 0 para recibir los segundos
  Wire.endTransmission(); //Termina la transmisión
  Wire.requestFrom(DS3231_I2C_ADDRESS, 7); //Se piden 7 bytes del RTC
  *second = bcdToDec(Wire.read() & 0x7f); //Read devuelve el último byte transmitido del RTC
  *minute = bcdToDec(Wire.read());        //Cuando el RTC recibe el OK de que el byte se ha mandado
  *hour = bcdToDec(Wire.read() & 0x3f);   //bien avanza el puntero hasta el siguiente y lo envía
  *dayOfWeek = bcdToDec(Wire.read());     //repitiendo el proceso para los 7 bytes.
  *dayOfMonth = bcdToDec(Wire.read());    //0x7f y 0x3f son filtros de dirección
  *month = bcdToDec(Wire.read());         //Como el RTC da los datos en formato BCD
  *year = bcdToDec(Wire.read());          //se pasan a decimal con la función bcdToDec
}

//// FUNCIÓN pagina0: programación de la página 0 de la interfaz de usuario,
// en la que se muestra por pantalla la hora y fecha actuales con formato:
// HORA -> 00:00    FECHA -> 00/00/0000 (día/mes/año)
void pagina0() {
  byte second, minute, hour, dayOfWeek, day, month, year;
  readDS3231time(&second, &minute, &hour, &dayOfWeek, &day, &month, &year);
  if (hour < 10) {
    myOLED.print("0", 47, 15); //Si la hora es <10, se añade un 0 delante para darle formato de 00:00
    myOLED.printNumI(hour, 53, 15);
  }
  else myOLED.printNumI(hour, 47, 15);
  myOLED.print(":", 59, 15);
  if (minute < 10) {
    myOLED.print("0", 65, 15);
    myOLED.printNumI(minute, 71, 15);
  }
  else myOLED.printNumI(minute, 65, 15);
  if (day < 10) {
    myOLED.print("0", 38, 30);
    myOLED.printNumI(day, 44, 30);
  }
  else myOLED.printNumI(day, 38, 30);
  myOLED.print("/", 50, 30);
  if (month < 10) {
    myOLED.print("0", 56, 30);
    myOLED.printNumI(month, 62, 30);
  }
  else myOLED.printNumI(month, 56, 30);
  myOLED.print("/", 68, 30);
  myOLED.printNumI(year, 74, 30);
}

//// FUNCIÓN pagina4: programación de la página 4 del menú, donde se leen
// los datos de humedad, temperatura y nivel de agua y se saca por pantalla
// un mensaje en función de las condiciones leídas. Además se muestra el
// tiempo que queda hasta el siguiente riego programado.
void pagina4() {
  byte second, minute, hour, dayOfWeek, day, month, year;
  readDS3231time(&second, &minute, &hour, &dayOfWeek, &day, &month, &year);
  sensorHum();
  nivelAgua();

  if (humedad <= 50) myOLED.print("Suelo seco.", 10, 15);
  else if (50 < humedad <= 80) myOLED.print("Suelo humedo.", 10, 15);
  else if (humedad > 80) myOLED.print("Exceso de humedad.", 10, 15);

  if (nivel <= 30) myOLED.print("Deposito vacio.", 10, 31);
  else if (30 < nivel <= 60) myOLED.print("Nivel bajo. Llenar.", 10, 31);
  else if (nivel > 60) myOLED.print("Deposito lleno.", 10, 31);

  if ((hour * 60 + minute > h_riego1 * 60 + m_riego1) && (hour * 60 + minute < h_riego2 * 60 + m_riego2))
    prox_riego = h_riego2 * 60 + m_riego2 - (hour * 60 + minute);
  else if (hour * 60 + minute > h_riego2 * 60 + m_riego2)
    prox_riego = 24 * 60 - (hour * 60 + minute) + (h_riego1 * 60 + m_riego1);
  else
    prox_riego = h_riego1 * 60 + m_riego1 - (hour * 60 + minute);

  myOLED.print("Riego en ", 23, 47);
  if (prox_riego % 60 < 10) {
    myOLED.print("0", 79, 47);
    myOLED.printNumI(prox_riego / 60, 85, 47);
  }
  else
    myOLED.printNumI(prox_riego / 60, 79, 47);
  myOLED.print(":", 91, 47);
  if (prox_riego % 60 < 10) {
    myOLED.print("0", 97, 47);
    myOLED.printNumI(prox_riego % 60, 103, 47);
  }
  else
    myOLED.printNumI(prox_riego % 60, 97, 47);
}

//// FUNCION ok: de tipo boolean, que devuelve un 1 o un 0 en función de
// si está pulsado el botón "OK" o no, respectivamente.
boolean ok() {
  int y = digitalRead(3);

  if (y == 0)
    selecOk = 0;
  if (y == 1 && selecOk == 0) {
    selecOk = 1;
    return true;
  }
  return false;
}

//// Función back: permite volver a la página anterior a la que se encuentra
// mediante el botón "atrás".
void back() {
  int z = digitalRead(4);
  if (z == 0)
    selecBack = 0;
  if (z == 1 && selecBack == 0) {
    i = 13;
    switch (page) {
      case 3:
        page = 1;
        break;
      default:
        if (page > 0)
          page = page - 1;
    }
    selecBack = 1;
  }
}

////FUNCIÓN down: permite navegar por el sistema mediante el botón "siguiente".
// n es el número de opciones del menú principal por las que se puede navegar.
void down(int n) {
  int x = digitalRead(2);

  if (x == 0)
    selecDown = 0;

  if (x == 1 && selecDown == 0) {
    selecDown = 1;
    myOLED.clrScr();
    myOLED.update();

    if (page == 1) {
      i = i + 14;
      if (i > 13 + (n - 1) * 14)
        i = 13;
    }
    else {
      i = i + 16;
      if (i > 15 + (n - 1) * 16)
        i = 15;
    }
  }
}

//// FUNCIÓN sleep: gestiona el encendido/apagado de la pantalla a través de un pulsador.
void sleep() {
  int w = digitalRead(5);

  if (w == 0)
    selecSleep = 0;

  if (w == 1 && selecSleep == 0) {
    selecSleep = 1;
    if (currentSleep == true) {
      myOLED.sleepMode(SLEEP_OFF);
      currentSleep = false;
    }
    else {
      myOLED.sleepMode(SLEEP_ON);
      currentSleep = true;
      page = 0;
    }
    myOLED.update();
  }
}

//// FUNCIÓN leerDHT: obtiene los valores de humedad y Tª del ambiente a través del sensor DHT11
void leerDHT() {
  humDHT = dht.readHumidity();
  temp = dht.readTemperature();
}

//// FUNCIÓN sensorHum: mide la humedad de la tierra a partir de los datos recogidos por los
// sensores de humedad resistivo y capacitivo, y hace una media de ambas medidas. Un LED
// amarillo indica que el nivel de humedad de la tierra es bajo y que se necesita regar.
void sensorHum() {
  int valorRes = 0;
  int valorCap = 0;
  valorRes = map(analogRead(sensorRes), 340, 1023, 100, 0);
  valorCap = map(analogRead(sensorCap), 200, 523, 100, 0);
  humedad = (valorRes + valorCap) / 2;
  if (humedad < 0) humedad = 0;
  if (humedad > 100) humedad = 100;
  if (humedad < 50)
    digitalWrite(yellLED , HIGH);
  else digitalWrite(yellLED , LOW);

}

//// FUNCIÓN nivelAgua: lee a través del sensor de nivel de agua la cantidad de agua que queda
// en el depósito. Cuando baja de un determinado nivel, se enciende un LED rojo para indicar
// que se necesita llenar.
void nivelAgua() {
  nivel = map(analogRead(sensorHW), 250, 580, 0, 100);
  if (nivel < 0) nivel = 0;
  if (nivel > 100) nivel = 100;
  if (nivel < 30)
    digitalWrite(redLED , HIGH);
  else digitalWrite(redLED , LOW);
}

//// FUNCIÓN servoDep: permite simular mediante un servomotor la apertura y el cierre del
// depósito de agua para poder llenarlo.
void servoDep() {
  if (i == 15 && page == 2 && ok()) {
    servo.write(0); //abrir depósito
  }
  else if (i == 31 && page == 2 && ok()) {
    servo.write(180); // cerrar depósito
  }
}

//// FUNCIÓN riegoManual: mediante la opción disponible en el menú de la pantalla de riego
// manual y su selección mediante el botón "OK" se activa el relé que permite funcionar a
// la bomba de agua y, por tanto, comenzar a regar. Este modo se para de la misma forma que
// se activó.
// Si se está regando, se enciende un LED verde.
void riegoManual() {
  if (i == 27 && page == 1 && ok()) {
    if (releON == false) {
      digitalWrite(bomba, LOW);
      digitalWrite(greenLED , HIGH);
      releON = true;
    }
    else {
      digitalWrite(bomba, HIGH);
      digitalWrite(greenLED , LOW);
      releON = false;
    }
  }
}

//// FUNCIÓN lecturas: realiza la lectura de datos cada 5 minutos de forma automática y
// actualiza el estado del sistema.
void lecturas() {
  currentMillis = millis();
  if ((currentMillis - previousMillis) > 300000) {
    previousMillis = currentMillis;
    leerDHT();
    sensorHum();
    nivelAgua();
  }
}

//// FUNCIÓN RiegoAuto: programa de riego automático, programado inicialmente para funcionar
// dos veces al día, que funciona solo si se cumplen una serie de condiciones en el sistema.
// También se ha programa el llenado del depósito de forma automática cuando baje de un
// determinado porcentaje.
// Si se está regando, se enciende un LED verde.
void RiegoAuto() {
  byte second, minute, hour, dayOfWeek, day, month, year;
  readDS3231time(&second, &minute, &hour, &dayOfWeek, &day, &month, &year);

  if (i == 13 && page == 1 && ok()) {
    if (autoON == false)
      autoON = true;
    else
      autoON = false;
  }

  if (autoON == true) {
    if (regando == true) {
      sensorHum();
      nivelAgua();
    }

    if ((hour == h_riego1 && minute == m_riego1 || hour == h_riego2 && minute == m_riego2) && ~regando ) {
      sensorHum();
      nivelAgua();
      if (humedad < 75 && nivel > 30) {
        digitalWrite(bomba, LOW);
        digitalWrite(greenLED , HIGH);
        regando = true;
        autoMillis = millis();
      }
    }

    if ((( millis() - autoMillis > 180000) && regando && humedad > 75) || nivel < 30) {
      digitalWrite(bomba, HIGH);
      digitalWrite(greenLED , LOW);
      regando = false;
    }

    if (nivel < 30)
      servo.write(0);
    if (nivel > 90)
      servo.write(180);
  }
}
